(define-module (ide completion)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (ice-9 poe)
  #:use-module (ice-9 pretty-print)
  #:use-module (ice-9 rdelim)
  #:use-module (ice-9 receive)
  #:use-module (gaiag parse)
  #:use-module (gaiag peg codegen)
  #:export (completions))

(define-public (command-line)
  '())

(module-define! (resolve-module '(gaiag command-line)) 'command-line command-line)

(define keywords
  '("behaviour" "bool" "component" "else" "enum" "false" "if" "illegal" "in"
    "inevitable" "interface" "namespace" "on" "optional" "out"
    "provides" "requires" "subint" "system" "true" "void"))

(define toplevel-declarations
  '("component" "enum" "interface" "namespace"))

(define (monotone-tree-collect monotone-predicate o)
  (if (monotone-predicate o)
      (cons o (append-map (cut monotone-tree-collect monotone-predicate <>) (cdr o)))
      '()))

(define (tree-collect predicate o)
  (cond ((predicate o) (list o))
        ((pair? o) (append-map (cut tree-collect predicate <>) (cdr o)))
        (else '())))

(define (interface-names o)
  (map (compose cadr cadr cadr)
       (tree-collect (lambda (o)
                       (and (pair? o)
                            (eq? (car o) 'interface)))
                     o)))

(define (port-names o)
  (map (compose second fourth)
       (tree-collect (lambda (o)
                       (and (pair? o)
                            (eq? (car o) 'port)))
                     o)))

(define (component-names o)
  (map (compose cadr cadr cadr)
       (tree-collect (lambda (o)
                       (and (pair? o)
                            (eq? (car o) 'component)))
                     o)))

(define (find-leaf o symbol)
  (find (lambda (o) (and (pair? o)
                         (eq? (car o) symbol)))
        (cdr o)))

(define (find-surrounding pt at)
  (receive (before after)
           (partition
            (lambda (o)
              (and (pair? o)
                   (let ((location (find-leaf o 'location)))
                     (and location
                          (< (third location) at)))))
            (cdr pt))
           (let ((after (filter (lambda (o) (and (pair? o)
                                                 (not (cut eq? 'location (car o)))))
                                after)))
            (cond ((and (null? before) (null? after)) '())
                  ((null? after) (list (last before)))
                  ((null? before) (list (first after)))
                  (else (list (last before) (first after)))))))

(define (find-leaf-at pt at)
  (reverse (monotone-tree-collect
            (lambda (o)
              (and (pair? o)
                   (let ((location (find-leaf o 'location)))
                     (and location
                          (<= (second location) at) (<= at (third location))))))
            pt)))

(define (find-nested pt)
  (find-leaf-at pt (third (find-leaf pt 'location))))

(define (find-context pt at)
  (let ((lst (find-leaf-at pt at)))
    (if (pair? lst) (append (find-surrounding (car lst) at) lst)
        lst)))

;; (let* ((pt (peg:parse ((compose read-string open-input-file) "test.dzn") ""))
;;        (context (find-context pt 43)))
;;   (pretty-print 'parse-tree)
;;   (pretty-print pt)
;;   (pretty-print 'context)
;;   (pretty-print context)
;;   (pretty-print 'nested)
;;   (pretty-print (find-nested (car context))))

;; (let ((pt (peg:parse ((compose read-string open-input-file) "test.dzn") "")))
;;   (pretty-print pt)
;;   (interface-names pt))

;;(define (message m) (warn m))

(unless (module-variable (current-module) 'message)
        (module-define! (current-module) 'message
                        (lambda (m) (warn m))))



;; highly incomplete and in need of a non brittle approach
(define (completions data at)
  (let* ((pt (parameterize ((%peg:fall-back? #t)) (peg:parse data "")))
         ;;(foo (pretty-print pt))
         (leaves (find-leaf-at pt at))
         (candidates (if (null? leaves) leaves
                         (find-surrounding (car leaves) at)))
         (context (if (null? candidates) leaves
                      (find-nested (car candidates)))))
    (message (format #f "caret position: ~a: context: ~a" at (map car context)))
    (if (pair? context)
        (cond
         ((eq? (caar context)
               'root)
          (string-join (sort toplevel-declarations string<?)))
         ((eq? (caar context)
               'interface)
          "in out")
         ((or (member (caar context)
                      '(ports port))
              (equal? (map car context) '(name compound-name)))
          "behaviour provides requires")
         ((and (member (caar context)
                       '(provides requires)))
          (string-join (delete-duplicates (sort (interface-names pt) string<?))))
         ((eq? (caar context) 'behaviour-compound)
          "on")
         ((equal? (map car context)
                  '(on behaviour-compound behaviour component root))
          (string-join (sort (port-names (fourth context)) string<?)))
         ((equal? (map car context)
                  '(on behaviour-compound behaviour component))
          "e")
         (else
          ""))
        "component interface namespace")))

;;(string-join (delete-duplicates (sort keywords string<?)))

;;(define completions (pure-funcq completions))
